@@ -1,8 +1,10 @@
 # flake8: noqa
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
 from .config import ConfigurationError
 from .config import DOCKER_CONFIG_KEYS
 from .config import find
-from .config import get_service_name_from_net
 from .config import load
 from .config import merge_environment
 from .config import parse_environment
